@@ -3,14 +3,16 @@
 # SPDX-License-Identifier: Apache-2.0
 """Instantion module."""
 from abc import ABC
-from dataclasses import dataclass
-from typing import Iterable, Optional
+from dataclasses import dataclass, field
+from typing import Any, Dict, Iterable, List, Optional
 from uuid import uuid4
 from warnings import warn
+
+from dacite import from_dict
+
 from onapsdk.exceptions import (
     APIError, InvalidResponse, ParameterError, ResourceNotFound, StatusError
 )
-
 from onapsdk.onap_service import OnapService
 from onapsdk.sdnc import NetworkPreload, VfModulePreload
 from onapsdk.sdc.service import Network, Service as SdcService, Vnf, VfModule
@@ -22,16 +24,63 @@ from .so_element import OrchestrationRequest
 
 
 @dataclass
-class VnfParameter:
-    """Class to store vnf parameter used for preload.
+class SoServiceVfModule:
+    """Class to store a VfModule instance parameters."""
 
-    IT'S DEPRECATED! `InstantiationParameter` SHOULD BE USED
+    model_name: str
+    instance_name: str
+    parameters: Dict[str, Any] = field(default_factory=dict)
+    processing_priority: Optional[int] = None
 
-    Contains two values: name of vnf parameter and it's value
+
+@dataclass
+class SoServiceXnf:
+    """Class to store a Xnf instance parameters."""
+
+    model_name: str
+    instance_name: str
+    parameters: Dict[str, Any] = field(default_factory=dict)
+    processing_priority: Optional[int] = None
+
+
+@dataclass
+class SoServiceVnf(SoServiceXnf):
+    """Class to store a Vnf instance parameters."""
+
+    vf_modules: List[SoServiceVfModule] = field(default_factory=list)
+
+
+@dataclass
+class SoServicePnf(SoServiceXnf):
+    """Class to store a Pnf instance parameters."""
+
+
+@dataclass
+class SoService:
+    """Class to store SO Service parameters used for macro instantiation.
+
+    Contains value list: List of vnfs to instantiate
+    Contains value: subscription service type
     """
 
-    name: str
-    value: str
+    subscription_service_type: str
+    vnfs: List[SoServiceVnf] = field(default_factory=list)
+    pnfs: List[SoServicePnf] = field(default_factory=list)
+    instance_name: Optional[str] = None
+
+    @classmethod
+    def load(cls, data: Dict[str, Any]) -> "SoService":
+        """Create a service instance description object from the dict.
+
+        Useful if you keep your instance data in file.
+
+        Returns:
+            SoService: SoService object created from the dictionary
+
+        """
+        return from_dict(data_class=cls, data=data)
+
+
 
 @dataclass
 class VnfParameters:
@@ -148,8 +197,8 @@ class VfModuleInstantiation(Instantiation):  # pytest: disable=too-many-ancestor
                               cloud_region: "CloudRegion" = None,
                               tenant: "Tenant" = None,
                               vf_module_instance_name: str = None,
-                              vnf_parameters: Iterable["InstantiationParameter"] = None
-                              ) -> "VfModuleInstantiation":
+                              vnf_parameters: Iterable["InstantiationParameter"] = None,
+                              use_preload: bool = True) -> "VfModuleInstantiation":
         """Instantiate VF module.
 
         Iterate throught vf modules from service Tosca file and instantiate vf modules.
@@ -172,22 +221,27 @@ class VfModuleInstantiation(Instantiation):  # pytest: disable=too-many-ancestor
                 "Python_ONAP_SDK_vf_module_service_instance_{str(uuid4())}".
                 Defaults to None.
             vnf_parameters (Iterable[InstantiationParameter], optional): Parameters which are
-                going to be used in preload upload for vf modules. Defaults to None.
+                going to be used in preload upload for vf modules or passed in "userParams".
+                Defaults to None.
+            use_preload (bool, optional): This flag determines whether instantiation parameters
+                are used as preload or "userParams" content. Defaults to True
 
         Yields:
             Iterator[VfModuleInstantiation]: VfModuleInstantiation class object.
 
         """
-        sdc_service: SdcService = vnf_instance.service_instance.service_subscription.sdc_service
         if vf_module_instance_name is None:
             vf_module_instance_name = \
                 f"Python_ONAP_SDK_vf_module_instance_{str(uuid4())}"
-        VfModulePreload.upload_vf_module_preload(
-            vnf_instance,
-            vf_module_instance_name,
-            vf_module,
-            vnf_parameters
-        )
+        if use_preload:
+            VfModulePreload.upload_vf_module_preload(
+                vnf_instance,
+                vf_module_instance_name,
+                vf_module,
+                vnf_parameters
+            )
+            vnf_parameters = None
+        sdc_service: SdcService = vnf_instance.service_instance.sdc_service
         response: dict = cls.send_message_json(
             "POST",
             (f"Instantiate {sdc_service.name} "
@@ -204,7 +258,8 @@ class VfModuleInstantiation(Instantiation):  # pytest: disable=too-many-ancestor
                     vnf_instance.service_instance.service_subscription.cloud_region,
                 tenant=tenant or \
                     vnf_instance.service_instance.service_subscription.tenant,
-                vnf_instance=vnf_instance
+                vnf_instance=vnf_instance,
+                vf_module_parameters=vnf_parameters or []
             ),
             headers=headers_so_creator(OnapService.headers)
         )
@@ -341,9 +396,12 @@ class VnfInstantiation(NodeTemplateInstantiation):  # pylint: disable=too-many-a
                               vnf_object: "Vnf",
                               line_of_business_object: "LineOfBusiness",
                               platform_object: "Platform",
-                              cloud_region: "CloudRegion" = None,
-                              tenant: "Tenant" = None,
-                              vnf_instance_name: str = None) -> "VnfInstantiation":
+                              cloud_region: "CloudRegion",
+                              tenant: "Tenant",
+                              sdc_service: "SdcService",
+                              vnf_instance_name: str = None,
+                              vnf_parameters: Iterable["InstantiationParameter"] = None
+                              ) -> "VnfInstantiation":
         """Instantiate Vnf using a'la carte method.
 
         Args:
@@ -359,18 +417,19 @@ class VnfInstantiation(NodeTemplateInstantiation):  # pylint: disable=too-many-a
                 THAT PROPERTY WILL BE REQUIRED IN ONE OF THE FUTURE RELEASE. REFACTOR YOUR CODE
                 TO USE IT!.
             vnf_instance_name (str, optional): Vnf instance name. Defaults to None.
+            vnf_parameters (Iterable[InstantiationParameter], optional): Instantiation parameters
+                that are sent in the request. Defaults to None
 
         Returns:
             VnfInstantiation: VnfInstantiation object
 
         """
-        sdc_service: SdcService = aai_service_instance.service_subscription.sdc_service
         if vnf_instance_name is None:
             vnf_instance_name = \
                 f"Python_ONAP_SDK_vnf_instance_{str(uuid4())}"
         response: dict = cls.send_message_json(
             "POST",
-            (f"Instantiate {aai_service_instance.service_subscription.sdc_service.name} "
+            (f"Instantiate {sdc_service.name} "
              f"service vnf {vnf_object.name}"),
             (f"{cls.base_url}/onap/so/infra/serviceInstantiation/{cls.api_version}/"
              f"serviceInstances/{aai_service_instance.instance_id}/vnfs"),
@@ -384,7 +443,8 @@ class VnfInstantiation(NodeTemplateInstantiation):  # pylint: disable=too-many-a
                 tenant=tenant or aai_service_instance.service_subscription.tenant,
                 line_of_business=line_of_business_object,
                 platform=platform_object,
-                service_instance=aai_service_instance
+                service_instance=aai_service_instance,
+                vnf_parameters=vnf_parameters or []
             ),
             headers=headers_so_creator(OnapService.headers)
         )
@@ -440,6 +500,7 @@ class ServiceInstantiation(Instantiation):  # pylint: disable=too-many-ancestors
                                  customer: "Customer",
                                  owning_entity: "OwningEntity",
                                  project: "Project",
+                                 service_subscription: "ServiceSubscription",
                                  service_instance_name: str = None) -> "ServiceInstantiation":
         """Instantiate service using SO a'la carte request.
 
@@ -460,7 +521,8 @@ class ServiceInstantiation(Instantiation):  # pylint: disable=too-many-ancestors
         """
         warn("This method is deprecated, use `instantiate_ala_carte`.", DeprecationWarning)
         return cls.instantiate_ala_carte(sdc_service, cloud_region, tenant, customer,
-                                         owning_entity, project, service_instance_name)
+                                         owning_entity, project, service_subscription,
+                                         service_instance_name)
 
     @classmethod
     def instantiate_ala_carte(cls,  # pylint: disable=too-many-arguments
@@ -470,6 +532,7 @@ class ServiceInstantiation(Instantiation):  # pylint: disable=too-many-ancestors
                               customer: "Customer",
                               owning_entity: "OwningEntity",
                               project: "Project",
+                              service_subscription: "ServiceSubscription",
                               service_instance_name: str = None,
                               enable_multicloud: bool = False) -> "ServiceInstantiation":
         """Instantiate service using SO a'la carte request.
@@ -481,6 +544,7 @@ class ServiceInstantiation(Instantiation):  # pylint: disable=too-many-ancestors
             customer (Customer): Customer to use in instantiation request
             owning_entity (OwningEntity): Owning entity to use in instantiation request
             project (Project): Project to use in instantiation request
+            service_subscription (ServiceSubscription): Customer's service subsription.
             service_instance_name (str, optional): Service instance name. Defaults to None.
             enable_multicloud (bool, optional): Determines if Multicloud should be enabled
                 for instantiation request. Defaults to False.
@@ -511,7 +575,8 @@ class ServiceInstantiation(Instantiation):  # pylint: disable=too-many-ancestors
                 owning_entity=owning_entity,
                 service_instance_name=service_instance_name,
                 project=project,
-                enable_multicloud=enable_multicloud
+                enable_multicloud=enable_multicloud,
+                service_subscription=service_subscription
             ),
             headers=headers_so_creator(OnapService.headers)
         )
@@ -527,8 +592,9 @@ class ServiceInstantiation(Instantiation):  # pylint: disable=too-many-ancestors
             project=project
         )
 
+    # pylint: disable=too-many-arguments, too-many-locals
     @classmethod
-    def instantiate_macro(cls,  # pylint: disable=too-many-arguments
+    def instantiate_macro(cls,
                           sdc_service: "SdcService",
                           customer: "Customer",
                           owning_entity: "OwningEntity",
@@ -540,7 +606,9 @@ class ServiceInstantiation(Instantiation):  # pylint: disable=too-many-ancestors
                           tenant: "Tenant" = None,
                           service_instance_name: str = None,
                           vnf_parameters: Iterable["VnfParameters"] = None,
-                          enable_multicloud: bool = False
+                          enable_multicloud: bool = False,
+                          so_service: "SoService" = None,
+                          service_subscription: "ServiceSubscription" = None
                           ) -> "ServiceInstantiation":
         """Instantiate service using SO macro request.
 
@@ -560,6 +628,9 @@ class ServiceInstantiation(Instantiation):  # pylint: disable=too-many-ancestors
                 going to be used for vnfs instantiation. Defaults to None.
             enable_multicloud (bool, optional): Determines if Multicloud should be enabled
                 for instantiation request. Defaults to False.
+            so_service (SoService, optional): SO values to use in instantiation request
+            service_subscription(ServiceSubscription, optional): Customer service subscription
+                for the instantiated service. Required if so_service is not provided.
 
         Raises:
             StatusError: if a service is not distributed.
@@ -568,34 +639,45 @@ class ServiceInstantiation(Instantiation):  # pylint: disable=too-many-ancestors
             ServiceInstantiation: instantiation request object
 
         """
+        template_file = "instantiate_service_macro.json.j2"
+        if so_service:
+            template_file = "instantiate_multi_vnf_service_macro.json.j2"
+            if so_service.instance_name:
+                service_instance_name = so_service.instance_name
+        else:
+            if not service_subscription:
+                raise ParameterError("If no so_service is provided, "
+                                     "service_subscription parameter is required!")
+            if service_instance_name is None:
+                service_instance_name = f"Python_ONAP_SDK_service_instance_{str(uuid4())}"
         if not sdc_service.distributed:
             msg = f"Service {sdc_service.name} is not distributed."
             raise StatusError(msg)
-        if service_instance_name is None:
-            service_instance_name = f"Python_ONAP_SDK_service_instance_{str(uuid4())}"
 
         response: dict = cls.send_message_json(
             "POST",
             f"Instantiate {sdc_service.name} service macro",
             (f"{cls.base_url}/onap/so/infra/"
              f"serviceInstantiation/{cls.api_version}/serviceInstances"),
-            data=jinja_env().get_template("instantiate_service_macro.json.j2").\
-            render(
-                sdc_service=sdc_service,
-                cloud_region=cloud_region,
-                tenant=tenant,
-                customer=customer,
-                owning_entity=owning_entity,
-                project=project,
-                aai_service=aai_service,
-                line_of_business=line_of_business,
-                platform=platform,
-                service_instance_name=service_instance_name,
-                vnf_parameters=vnf_parameters,
-                enable_multicloud=enable_multicloud
-            ),
+            data=jinja_env().get_template(template_file). \
+                render(
+                    so_service=so_service,
+                    sdc_service=sdc_service,
+                    cloud_region=cloud_region,
+                    tenant=tenant,
+                    customer=customer,
+                    owning_entity=owning_entity,
+                    project=project,
+                    aai_service=aai_service,
+                    line_of_business=line_of_business,
+                    platform=platform,
+                    service_instance_name=service_instance_name,
+                    vnf_parameters=vnf_parameters,
+                    enable_multicloud=enable_multicloud,
+                    service_subscription=service_subscription
+                ),
             headers=headers_so_creator(OnapService.headers)
-        )
+            )
         return cls(
             name=service_instance_name,
             request_id=response["requestReferences"].get("requestId"),
@@ -689,7 +771,6 @@ class NetworkInstantiation(NodeTemplateInstantiation):  # pylint: disable=too-ma
             NetworkInstantiation: NetworkInstantiation object
 
         """
-        sdc_service: SdcService = aai_service_instance.service_subscription.sdc_service
         if network_instance_name is None:
             network_instance_name = \
                 f"Python_ONAP_SDK_network_instance_{str(uuid4())}"
@@ -698,7 +779,7 @@ class NetworkInstantiation(NodeTemplateInstantiation):  # pylint: disable=too-ma
                                               subnets=subnets)
         response: dict = cls.send_message_json(
             "POST",
-            (f"Instantiate {aai_service_instance.service_subscription.sdc_service.name} "
+            (f"Instantiate {aai_service_instance.sdc_service.name} "
              f"service network {network_object.name}"),
             (f"{cls.base_url}/onap/so/infra/serviceInstantiation/{cls.api_version}/"
              f"serviceInstances/{aai_service_instance.instance_id}/networks"),
@@ -706,7 +787,7 @@ class NetworkInstantiation(NodeTemplateInstantiation):  # pylint: disable=too-ma
             render(
                 instance_name=network_instance_name,
                 network=network_object,
-                service=sdc_service,
+                service=aai_service_instance.sdc_service,
                 cloud_region=cloud_region or \
                     aai_service_instance.service_subscription.cloud_region,
                 tenant=tenant or aai_service_instance.service_subscription.tenant,
